/*
 * Copyright (c) 2005 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


using Nemerle.Collections;
using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;


namespace Nemerle
{
  [Nemerle.MacroUsage (Nemerle.MacroPhase.WithTypedMembers,
                       Nemerle.MacroTargets.Method)]
  macro Memoize (tb : TypeBuilder, meth : MethodBuilder, params opts : list [PExpr])
  {
    def _is_static = meth.IsStatic;
    mutable invalid = null;
    mutable invalid_is_null = false;
    mutable synch = true;

    foreach (o in opts) {
      | <[ InvalidValue = null ]> =>
        invalid = <[ null ]>;
        invalid_is_null = true;

      | <[ InvalidValue = $expr ]> =>
        when (invalid != null) Message.Error ("invalid value specified more than once");
        invalid = expr;

      | <[ Synchronized = $(x : bool) ]> =>
        synch = x;

      | _ => Message.Error ($"unrecognized parameter: $o")
    }

    def store = Macros.NewSymbol ("storage");
    def is_cached =
      if (invalid == null) {
        def x = Macros.NewSymbol ("is_cached");
        tb.Define (<[ decl: private mutable $(x : name) : bool; ]>);
        <[ $(x : name) ]>
      }
      else
        <[ $(store : name) != $invalid ]>;

    if (invalid != null && ! invalid_is_null)
      tb.Define (<[ decl: private mutable $(store : name) : $(meth.ReturnType : typed) = $invalid; ]>);
    else
      tb.Define (<[ decl: private mutable $(store : name) : $(meth.ReturnType : typed); ]>);

    meth.Body = <[
      if ($is_cached)
        $(store : name)
      else {
        $(store : name) = $(meth.Body);
        $(if (invalid == null) <[ $is_cached = true ]> else <[ () ]>);
        $(store : name)
      }
    ]>
  }
}
