/*
 * Copyright (c) 2003, 2004 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

namespace Sioux
{
    using Nemerle.Collections;
  
    using System.Security.Cryptography;
    using System;

    /**
     *  Module for computing session ids
     */
    internal module SessionID
    {
       
        /* -- INTERNAL METHODS -------------------------------------------------------- */

        internal Initialize() : void
        {
            allowed =  array ['0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'];
            id_length = 30;
        }
        
        /**
         *  Computes unique session id
         */
        internal GetSessionID () : string
        {
            GetSessionID(id_length)            
        }

        internal GetSessionID (len : int) : string
        {
            def random = array (len/2);
            def id     = array (len);
            def rng    = RNGCryptoServiceProvider ();
            
            rng.GetBytes(random);
       
            for(mutable i = 0;i < len/2; i=i+1)
            {
                id [i * 2] = allowed [random[i] >> 4];
                id [(i * 2) + 1] = allowed [(random[i] :> int) % 16];                
            }

            String(id)
        }
        
        /* -- PRIVATE FIELDS ------------------------------------------------------- */

        private mutable allowed : array[char];
        private mutable id_length : int;
    }
}
